/*
 * dumpici.c - Obtain an ICI and pretty-print its content
 *
 * Written 2002 by Werner Almesberger
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netinet/in.h>

#include <linux/tcpcp.h>

#include <tcpcp.h>

#include "util.h"


static int verbose = 0; /* dump buffer data */


#define IP_QUAD(ip) \
  ((uint8_t *) &(ip))[0],((uint8_t *) &(ip))[1], \
  ((uint8_t *) &(ip))[2], ((uint8_t *) &(ip))[3]


static void print_ici_v(const struct tcpcp_icie_version *v)
{
    unsigned flags;
    int i;

    printf("  major:\t\t%u\n",v->major);
    printf("  minor:\t\t%u\n",v->minor);
    printf("  ip_version:\t\t%u\n",v->ip_version);
    flags = ntohs(v->flags);
    printf("  flags:\t\t0x%x\n",flags);

    if (flags & TCPCP_ICIF_USEPERF) printf("    TCPCP_ICIF_USEPERF\n");
    for (i = 4; i <= 32768; i += i)
	if (flags & i) printf("    ??? (%d)\n",i);
    printf("  ici_hdr_size:\t\t%lu\n",(unsigned long) ntohs(v->ici_hdr_size));
    printf("  buf_hdr_size:\t\t%lu\n",(unsigned long) ntohs(v->buf_hdr_size));
}


static void print_ici_id(const struct tcpcp_icie_id *id)
{
    printf("  ip.v4.ip_src:\t\t%u.%u.%u.%u\n",IP_QUAD(id->ip.v4.ip_src));
    printf("  ip.v4.ip_dst:\t\t%u.%u.%u.%u\n",IP_QUAD(id->ip.v4.ip_dst));
    printf("  tcp_sport:\t\t%lu\n",(unsigned long) ntohs(id->tcp_sport));
    printf("  tcp_dport:\t\t%lu\n",(unsigned long) ntohs(id->tcp_dport));
}


static void print_ici_fixgen(const struct tcpcp_icie_fixgen *fix)
{
    printf("  tcp_flags:\t\t0x%x\n",fix->tcp_flags);
    /* can't include linux/tcp.h :-( */
    if (fix->tcp_flags & 1) printf("    TCPI_OPT_TIMESTAMPS\n");
    if (fix->tcp_flags & 2) printf("    TCPI_OPT_SACK\n");
    if (fix->tcp_flags & 4) printf("    TCPI_OPT_WSCALE\n");
    if (fix->tcp_flags & 8) printf("    TCPI_OPT_ECN\n");
    if (fix->tcp_flags & 16) printf("    ??? (16)\n");
    if (fix->tcp_flags & 32) printf("    ??? (32)\n");
    if (fix->tcp_flags & 64) printf("    ??? (64)\n");
    if (fix->tcp_flags & 128) printf("    ??? (128)\n");
    printf("  snd_wscale:\t\t%lu\n",(unsigned long) fix->snd_wscale);
    printf("  rcv_wscale:\t\t%lu\n",(unsigned long) fix->rcv_wscale);
    printf("  snd_mss:\t\t%lu\n",(unsigned long) ntohs(fix->snd_mss));
    printf("  rcv_mss:\t\t%lu\n",(unsigned long) ntohs(fix->rcv_mss));
}


static void print_ici_vargen(const struct tcpcp_icie_vargen *var)
{
    static const char *state[] = {
	"???", "TCP_ESTABLISHED", "TCP_SYN_SENT", "TCP_SYN_RECV",
	"TCP_FIN_WAIT1", "TCP_FIN_WAIT2", "TCP_TIME_WAIT", "TCP_CLOSE",
	"TCP_CLOSE_WAIT", "TCP_LAST_ACK", "TCP_LISTEN", "TCP_CLOSING",
    };

    printf("  state:\t\t");
    if (var->state < sizeof(state)/sizeof(const char *))
	printf("%s",state[var->state]);
    else printf("???");
    printf(" (%u)\n",var->state);
    printf("  snd_nxt:\t\t%lu\n",(unsigned long) ntohl(var->snd_nxt));
    printf("  rcv_nxt:\t\t%lu\n",(unsigned long) ntohl(var->rcv_nxt));
    printf("  snd_wnd:\t\t%lu\n",(unsigned long) ntohl(var->snd_wnd));
    printf("  rcv_wnd:\t\t%lu\n",(unsigned long) ntohl(var->rcv_wnd));
    printf("  ts_gen:\t\t%lu\n",(unsigned long) ntohl(var->ts_gen));
    printf("  ts_recent:\t\t%lu\n",(unsigned long) ntohl(var->ts_recent));
}


static void print_ici_stat(const struct tcpcp_icie_stat *st)
{
}


static void print_bufs(const struct tcpcp_icie_buf *buf,int buf_hdr_size,
  const void *end)
{
    unsigned length;
    int i;

    while (buf != end) {
	if ((char *) buf+buf_hdr_size > (char *) end) {
	    fprintf(stderr,"buffer header exceeds ICI\n");
	    exit(1);
	}
	length = ntohs(buf->length);
	switch (buf->type) {
	    case TPCPC_ICIE_BUF_SND:
		printf("send");
		break;
	    case TPCPC_ICIE_BUF_OOO:
		printf("out-of-order (receive)");
		break;
	    default:
		printf("unknown (%u)",buf->type);
		break;
	}
	if ((void *) (buf->data+length) > end) {
	    fflush(stdout);
	    fprintf(stderr,"\nbuffer (length %u) exceeds ICI\n",length);
	    exit(1);
	}
	printf(" buffer: %u bytes, seq %lu\n",length,
	    (unsigned long) ntohl(buf->seq));
	if (verbose)
	    for (i = 0; i < length; i += 16) {
		int end,j;

		printf("  %04x ",i);
		end = i+15 < length ? i+16 : length;
		for (j = i; j < end; j++) {
		    if (!(j & 7)) putchar(' ');
		    printf("%02x ",buf->data[j]);
		}
		putchar(' ');
		for (j = i; j < end; j++) {
		    uint8_t byte = buf->data[j];

		    putchar(byte < ' ' || byte > '~' ? '.' : byte);
		}
		putchar('\n');
	    }
	buf = (struct tcpcp_icie_buf *)
	  ((char *) buf+buf_hdr_size+((length+3) & ~3));
    }
}


static void print_ici(const struct tcpcp_ici *ici,int size)
{
    printf("ICI: loaded %d, ",size);
    if (size == ntohl(ici->ici_length)) printf("as expected\n");
    else printf("expected %lu\n",(unsigned long) ntohl(ici->ici_length));
    printf("version:\n");
    print_ici_v(&ici->v);
    printf("id:\n");
    print_ici_id(&ici->id);
    printf("fixgen:\n");
    print_ici_fixgen(&ici->fixgen);
    printf("vargen:\n");
    print_ici_vargen(&ici->vargen);
    printf("stat:\n");
    print_ici_stat(&ici->stat);
    print_bufs((const struct tcpcp_icie_buf *) ici->buf,
      ntohs(ici->v.buf_hdr_size),(const char *) ici+size);
}


static void print_short(const struct tcpcp_ici *ici,int size)
{
    const struct tcpcp_icie_id *id = &ici->id;

    printf("%u.%u.%u.%u %lu\n",IP_QUAD(id->ip.v4.ip_src),
      (unsigned long) ntohs(id->tcp_sport));
    printf("%u.%u.%u.%u %lu\n",IP_QUAD(id->ip.v4.ip_dst),
      (unsigned long) ntohs(id->tcp_dport));
}


static void check_ici(const struct tcpcp_ici *ici,int size)
{
    if (size < 4) {
	fprintf(stderr,"minimum ICI size is 4, not %d\n",size);
	exit(1);
    }
    if (ntohs(ici->v.ici_hdr_size) > size) {
	fprintf(stderr,"loaded ICI (%d bytes) is smaller than fixed size %u\n",
	  size,ntohs(ici->v.ici_hdr_size));
	exit(1);
    }
}


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s [-v] [ici_file]\n\n",name);
    fprintf(stderr,"  -s  short output (only print local/remote "
      "address/port)\n");
    fprintf(stderr,"  -v  verbose mode (dump data buffer content)\n");
    exit(1);
}


int main(int argc,char **argv)
{
    struct tcpcp_ici *ici;
    char c;
    int fd = 0,short_output = 0,size;

    while ((c = getopt(argc,argv,"sv")) != EOF)
	switch (c) {
	    case 's':
		short_output = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    default:
		usage(argv[0]);
	}
    if (optind+1 < argc) usage(argv[0]);
    if (optind+1 == argc) {
	fd = open(argv[optind],O_RDONLY);
	if (fd < 0) {
	    perror(argv[optind]);
	    return 1;
	}
    }
    ici = read_file(fd,&size);
    check_ici(ici,size);
    if (short_output) print_short(ici,size);
    else print_ici(ici,size);
    return 0;
}
