/*
 * sendcp.c - Example HTTP sender using connection passing
 *
 * Written 2002 by Werner Almesberger
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <tcpcp.h>

#include "util.h"


/* ----- Accept new connection --------------------------------------------- */


static int new_connection(int port)
{
    int s;
    int state = 0;

    s = accept_connection(port);
    while (1) {
	char buf[1024];
	int got,i;

	got = read(s,buf,sizeof(buf));
	if (got < 0) {
	    perror("read");
	    exit(1);
	}
	if (!got) {
	    fprintf(stderr,"unexpected EOF on socket\n");
	    exit(1);
	}
	for (i = 0; i < got; i++) {
	    if (buf[i] == "\r\n\r\n"[state]) state++;
	    else state = 0;
	    if (state == 4) {
		write_file(s,"HTTP/1.0 200\r\n\r\n",16);
		return s;
	    }
	}
    }
}



/* ----- File IO ----------------------------------------------------------- */


static int open_file(const char *name,off_t offset)
{
    int fd;
    struct stat st;

    fd = open(name,O_RDONLY);
    if (fd < 0) {
	perror(name);
	exit(1);
    }
    if (fstat(fd,&st) < 0) {
	perror("stat");
	exit(1);
    }
    if (st.st_size < offset) {
	fprintf(stderr,"offset beyond end of file\n");
	exit(1);
    }
    if (lseek(fd,offset,SEEK_SET) == (off_t) -1) {
	perror("lseek");
	exit(1);
    }
    return fd;
}


static void send_data(int fd,int s,size_t bytes)
{
    char buf[8192];

    while (bytes) {
	ssize_t got;

	got = read(fd,buf,sizeof(buf) < bytes ? sizeof(buf) : bytes);
	if (got < 0) {
	    perror("read");
	    exit(1);
	}
	if (!got) exit(0);
	write_file(s,buf,got);
	bytes -= got;
    }
}


/* ----- Command-line handling --------------------------------------------- */


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s [-e] [-l data_port] [-r position] [-t position] "
      "ici_file file\n\n",name);
    fprintf(stderr,"  -e  exit immediately after dumping, without waiting for "
      "a signal\n");
    fprintf(stderr,"  -l  listen for incoming HTTP connection on "
      "\"data_port\"\n");
    fprintf(stderr,"  -r  resume at absolute file position \"position\"\n");
    fprintf(stderr,"  -t  stop at file byte \"position\", with exit code 2\n");
    exit(1);
}


int main(int argc,char **argv)
{
    unsigned long listen_port = 0,resume_at = 0,stop_at = ~0UL;
    int do_wait = 1;
    char *end;
    char c,fd,s;

    while ((c = getopt(argc,argv,"el:r:t:")) != EOF)
	switch (c) {
	    case 'e':
		do_wait = 0;
		break;
	    case 'l':
		listen_port = strtoul(optarg,&end,10);
		if (*end || !listen_port || (listen_port & ~0xffff))
		    usage(argv[0]);
		break;
	    case 'r':
		resume_at = strtoul(optarg,&end,10);
		if (*end) usage(argv[0]);
		break;
	    case 't':
		stop_at = strtoul(optarg,&end,10);
		if (*end) usage(argv[0]);
		break;
	    default:
		usage(argv[0]);
	}
    if (optind+2 != argc) usage(argv[0]);
    if (stop_at < resume_at) {
	fprintf(stderr,"stop_at < resume_at\n");
	exit(1);
    }
    fd = open_file(argv[optind+1],resume_at);
    if (listen_port) s = new_connection(listen_port);
    else s = resurrect_connection(argv[optind]);
    send_data(fd,s,stop_at-resume_at);
    dump_ici(s,argv[optind],1);
    if (do_wait) {
	fprintf(stderr,"Dumped. Waiting for signal.\n");
	fflush(stderr);
	while (1) pause();
    }
    return 0;
}
