/*
 * shmod.c - Shell module
 *
 * Written 1997,2000-2002 by Werner Almesberger
 *
 * Based on ush (from Linux-7k), bsh (from newlib-linux), and a bit of rbsh
 * (for BB).
 */


/* uh, need to revise comments */

/*
 * Note: this shell does NOT try to conform to POSIX 1003.2, although some
 * of its parsing rules are similar. The main purpose of this shell is to
 * offer enough functionality for a luxurious boot prompt while still not
 * wasting too much space.
 */

/*
 * TODO:
 * - maintain word boundaries in double quoted strings (e.g. "* x*")
 * - backtick expansion
 * - add io redirection ? | ? || ? && ?
 * - check for buffer overflows (variable(?))
 * - set last_status for internal commands
 * - catch signals (or, rather, don't generate them)
 * - limited command-line editing with file name completion
 * - print signal names instead of numbers
 * - write regression test
 * - add some command line options
 * - clean up globbing
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "util.h"


#define RC_FILE	"/etc/bshrc"

#define CWD_BUF		4096
#define LINE_BUF	4096
#define MAX_ARGS	1024
#define MAX_WORD	2048
#define IN_STACK	  16	/* NB: unlike sh, bsh can recurse here */


int shmod_forced_tty = -1;
int shmod_catch_signals = 1;


static void wait_background(void);
static int dont_expand = 0; /* prevent x=\$x; echo $x loop */


/* ----- Character input --------------------------------------------------- */


struct input {
    int (*next_char)(int new_command,void **user);
    void *user;
} in_stack[IN_STACK],*in = in_stack;


static void push_input(int (*fn)(int new_command,void **user),void *user)
{
    if (in-in_stack == IN_STACK-1) {
	fprintf(stderr,"input stack overflow\n");
	return;
    }
    in++;
    in->next_char = fn;
    in->user = user;
}


static void pop_input(void)
{
    in--;
}


static int next_char(int new_command);


static void show_prompt(int full_prompt)
{
    char cwd_buf[CWD_BUF];

    wait_background();
    if (!full_prompt) {
	printf("> ");
	fflush(stdout);
	return;
    }
    if (!getcwd(cwd_buf,sizeof(cwd_buf))) {
	perror("getcwd");
	return;
    }
    printf("%s! ",cwd_buf);
    fflush(stdout);
}


static int in_file(int new_command,void **user)
{
    static char line_buf[LINE_BUF];
    static const char *pos = line_buf;

    if (!*pos) {
	char *curr = line_buf;

	pos = line_buf;
	*line_buf = 0; /* for ^C */
	while (1) {
	    int escape;

	    if (isatty(fileno((FILE *) *user)) ||
	      fileno((FILE *) *user) == shmod_forced_tty)
		show_prompt(new_command);
	    if (!fgets(curr,sizeof(line_buf)-(curr-line_buf),*user)) {
		if (curr == line_buf) {
		    pos = line_buf; /* we may return to another file */
		    line_buf[0] = 0;
		    (void) fclose(*user);
		    pop_input();
		    return next_char(new_command);
		}
		break;
	    }
	    for (escape = 0; *curr; curr++)
		if (!escape) escape = *curr == '\\';
		else {
		    escape = 0;
		    if (curr[0] == '\n' && !curr[1]) {
			curr++; /* need to preserve \\n because we can only
				   figure out what needs to be done later
				   (consider \\n vs. '\\n') */
			break;
		    }
		}
	    if (!escape) break;
	    new_command = 0;
	}
	pos = line_buf;
    }
    return *pos++;
}


static int in_string(int new_command,void **user)
{
    if (*user) {
	char ch = *(char *) *user;
	if (ch) {
	    (*(char **) user)++;
	    return ch;
	}
    }
    dont_expand = 0;
    pop_input();
    return next_char(new_command);
}


static int in_char(int new_command,void **user)
{
    pop_input();
    return (int) *user;
}


static int in_eof(int new_command,void **user)
{
    return EOF;
}


static int next_char(int new_command)
{
    return in->next_char(new_command,&in->user);
}


static void unwind_stack(void)
{
    /* @@@ leak */
    if (in-in_stack > 2) in = in_stack+2;
}


static void return_char(int ch)
{
    push_input(&in_char,(void *) ch);
}


/* ----- Program execution ------------------------------------------------- */


static int last_status = 0;


static void wait_background(void)
{
    pid_t pid;
    int status;

    pid = waitpid(-1,&status,WNOHANG);
    if (pid < 0) {
	if (errno != ECHILD) perror("waitpid");
	return;
    }
    if (WIFEXITED(status)) printf("%d  Done\n",pid);
    if (WIFSIGNALED(status))
	printf("%d  Signal %d\n",pid,WTERMSIG(status));
}


static void wait_foreground(pid_t pid)
{
    int status;

    if (waitpid(pid,&status,0) < 0) {
	perror("waitpid");
	return;
    }
    last_status = WEXITSTATUS(status);
    if (WIFSIGNALED(status)) printf("Signal %d\n",WTERMSIG(status));
}


static void run_external(char **argv,int bg)
{
    pid_t pid;

    pid = fork();
    if (pid < 0) {
	perror("fork");
	return;
    }
    if (pid) {
	if (bg) {
	    printf("%d\n",pid);
	    return;
	}
	wait_foreground(pid);
	return;
    }
    execvp(*argv,argv);
    perror(*argv);
    _exit(0);
}


static int run_internal(char **argv)
{
    if (!strcmp(*argv,"cd")) {
	char *dir;

	dir = argv[1] ? argv[1] : getenv("HOME");
	if (!dir) dir = "/";
	if (chdir(dir)) perror("chdir");
	return 0;
    }
    if (strchr(*argv,'=')) {
	if (putenv(*argv)) perror("putenv");
	return 0;
    }
    if (!strcmp(*argv,".")) {
	FILE *file;

	if (!argv[1]) {
	    fprintf(stderr,"file name required\n");
	    return 0;
	}
	file = fopen(argv[1],"r");
	if (file) push_input(&in_file,file);
	else perror(argv[1]);
	return 0;
    }
    if (!strcmp(*argv,"echo")) {
	for (++argv; *argv; argv++)
	    printf("%s%s",*argv,argv[1] ? " " : "");
	putchar('\n');
	return 0;
    }
    if (!strcmp(*argv,"exec")) {
	if (argv[1]) {
	    execvp(argv[1],argv+1);
	    perror(argv[1]);
	}
	return 0;
    }
    if (!strcmp(*argv,"exit")) {
	if (argv[1]) last_status = atoi(argv[1]);
	exit(last_status);
    }
    return -1;
}


static struct cmd {
    const char *name;
    int (*fn)(char **argv);
    struct cmd *next;
} *added_cmds = NULL;


static int run_added(char **argv)
{
    const struct cmd *cmd;

    for (cmd = added_cmds; cmd; cmd = cmd->next)
	if (!strcmp(cmd->name,*argv)) {
	    last_status = cmd->fn(argv);
	    return 0;
	}
    return -1;
}


static void run_command(int argc,char **argv,int bg)
{
#if 0
    int i;

    for (i = 0; i < argc; i++) printf("%d: \"%s\"\n",i,argv[i]);
#endif
    if (!argc) return;
    argv[argc] = NULL;
    if (!run_internal(argv)) return;
    if (!run_added(argv)) return;
    run_external(argv,bg);
}


void shmod_add_command(const char *name,int (*fn)(char **argv))
{
    struct cmd *cmd;

    cmd = malloc(sizeof(*cmd));
    if (!cmd) {
	perror("malloc");
	exit(1);
    }
    cmd->name = strdup(name);
    if (!cmd->name) {
	perror("strdup");
	exit(1);
    }
    cmd->fn = fn;
    cmd->next = added_cmds;
    added_cmds = cmd;
}


/* ----- File name globbing ------------------------------------------------ */


#define GLOB_ASTERISK	'\001'
#define GLOB_QMARK	'\002'


static char glob_buf[MAX_WORD];


/*
 * Ugly and slow ...
 */

static int glob_match(const char *pat,const char *s)
{
    if (!*pat && !*s) return 1;
    switch (*pat) {
	case GLOB_ASTERISK:
	    while (1) {
		if (glob_match(pat+1,s)) return 1;
		if (!*s) break;
		s++;
	    }
	    return 0;
	default:
	    if (*pat != *s) return 0;
	    /* fall through */
	case GLOB_QMARK:
	    if (!*pat || !*s) return 0;
	    return glob_match(pat+1,s+1);
    }
}


static int glob_elem(const char *word,char *here,int *argc,char **argv,
  int root);


static int glob_item(const char *word,char *here,int *argc,char **argv)
{
    char *next;
    struct stat st;

    if (stat(glob_buf,&st)) return 0;
    if (!word) {
	if (*argc == MAX_ARGS-1) return -1;
	argv[(*argc)++] = strdup(glob_buf);
	return 1;
    }
    next = strchr(here,0);
    return glob_elem(word,next,argc,argv,0);
}


static int glob_elem(const char *word,char *here,int *argc,char **argv,int root)
{
    DIR *dir;
    struct dirent *de;
    char *slash;
    int entries;

    while (1) {
	slash = strchr(word,'/');
	if (slash) *slash++ = 0;
	if (root == 1 || (here != glob_buf && root != -1)) *here++ = '/';
	if (root) root = -1;
	*here = 0;
	if (*word || !slash) break;
	word = slash;
    }
    if (!root && !strchr(word,GLOB_ASTERISK) && !strchr(word,GLOB_QMARK)) {
	if (here-glob_buf+strlen(word) >= MAX_WORD-1) return -1;
	strcpy(here,word);
	return glob_item(slash,here,argc,argv);
    }
    entries = 0;
    dir = opendir(*glob_buf ? glob_buf : ".");
    if (!dir) return 0;
    while (1) {
	de = readdir(dir);
	if (!de) break;
	if (*de->d_name == '.') continue;
	if (glob_match(word,de->d_name)) {
	    int res;

	    if (here-glob_buf+strlen(de->d_name) >= MAX_WORD-1) return -1;
	    strcpy(here,de->d_name);
	    res = glob_item(slash,here,argc,argv);
	    if (res < 0) return res;
	    entries += res;
	}
    }
    if (closedir(dir)) perror("closedir");
    return entries;
}


static int do_glob_word(char *word,int *argc,char **argv)
{
    char *p;
    int res;

    if (*argc == MAX_ARGS-1) return -1;
    if (!strchr(word,GLOB_ASTERISK) && !strchr(word,GLOB_QMARK)) {
	argv[(*argc)++] = strdup(word);
	return 0;
    }
    res = glob_elem(word,glob_buf,argc,argv,*word == '/');
    if (res) return res;
    for (p = word; *p; p++)
	if (*p == GLOB_ASTERISK) *p = '*';
	else if (*p == GLOB_ASTERISK) *p = '?';
    argv[(*argc)++] = strdup(word);
    return 0;
}


static int glob_word(char *word,int *argc,char **argv)
{
    if (do_glob_word(word,argc,argv) >= 0) return 0;
    fprintf(stderr,"too many arguments\n");
    return -1;
}


/* ----- Variable expansion ------------------------------------------------ */


static char *lookup_var(const char *name)
{
    static char var_buf[20]; /* $$, $? */

    if (!name[1])
	switch (*name) {
	    case '$':
		sprintf(var_buf,"%d",getpid());
		return var_buf;
	    case '?':
		sprintf(var_buf,"%d",last_status);
		return var_buf;
	    default:
		break;
	}
    return getenv(name);
}


static int expand_var(void)
{
    char var_name[MAX_WORD];
    char *pos = var_name;
    int ch;

    ch = next_char(0);
    if (ch == '{') {
	while (1) {
	    ch = next_char(0);
	    if (ch == EOF) {
		fprintf(stderr,"syntax error\n");
		return -1;
	    }
	    if (ch == '}') break;
	    if (pos == var_name+MAX_WORD-1) {
		fprintf(stderr,"too long\n");
		return -1;
	    }
	    *pos++ = ch;
	}
    }
    else {
	while (ch != EOF && (isalnum(ch) || ch == '_' ||
	  (pos == var_name && strchr("$?",ch)))) {
	    if (pos == var_name+MAX_WORD-1) {
		fprintf(stderr,"too long\n");
		return -1;
	    }
	    *pos++ = ch;
	    ch = next_char(0);
	}
	return_char(ch);
    }
    *pos = 0;
    dont_expand = 1; /* @@@ need to revise this for backtick evaluation */
    push_input(&in_string,lookup_var(var_name));
    return 0;
}


/* ----- Command line parsing ---------------------------------------------- */


static void do_handle_command(void)
{
    char *argv[MAX_ARGS];
    char word[MAX_WORD];
    char *pos;
    int argc,escape,backtick,in_word;
    int quote = 0; /* 1: '; 2: " */

reset:
    pos = word;
    argc = escape = quote = backtick = in_word = 0;
    while (1) {
	int ch;

	ch = next_char(!argc);
	if (ch == EOF) {
	    if (escape || quote || backtick) fprintf(stderr,"syntax error\n");
	    else {
		if (in_word) {
		    in_word = 0;
		    *pos = 0;
		    if (glob_word(word,&argc,argv) < 0) goto reset;
		}
		run_command(argc,argv,0);
            }
	    exit(last_status);
	}
	if (escape) {
	    escape = 0;
	    /* slight difference to /bin/sh: "\a" -> a, not \a */
	    if (ch != '\n') goto append;
	    continue;
	}
	if (quote == 1) {
	    if (ch != '\'') goto append;
	    quote = 0;
	    continue;
	}
	switch (ch) {
	    case '\'':
		if (quote || backtick) break;
		quote = in_word = 1;
		continue;
	    case '"':
		if (backtick) break;
		quote = 2-quote; /* 0 or 2 */
		if (quote) in_word = 1;
		continue;
	    case '`':
		backtick = !backtick;
		if (backtick) in_word = 1;
		continue;
	    case '$':
		if (dont_expand) goto append;
		if (expand_var()) goto reset;
		continue;
	    case '\\':
		escape = in_word = 1;
		continue;
	    case '*':
		*pos++ = GLOB_ASTERISK;
		in_word = 1;
		continue;
	    case '?':
		*pos++ = GLOB_QMARK;
		in_word = 1;
		continue;
	    default:
		break;
	}
	if (quote || backtick) goto append;
	switch (ch) {
	    case ' ':
	    case '\t':
		if (!in_word) break;
		in_word = 0;
		*pos = 0;
		if (glob_word(word,&argc,argv) < 0) goto reset;
		pos = word;
		continue;
	    case '#':
		if (in_word) goto append;
		do ch = next_char(!argc);
		while (ch != '\n' && ch != EOF);
		/* fall through */
	    case ';':
	    case '\n':
	    case '&':
		if (in_word) {
		    in_word = 0;
		    *pos = 0;
		    if (glob_word(word,&argc,argv) < 0) goto reset;
		}
		run_command(argc,argv,ch == '&');
		goto reset;
	    default:
		if (ch < ' ' || ch > '~') continue;
		in_word = 1;
		goto append;
	}
	continue;

append:
	if (pos-word == MAX_WORD-1) {
	    fprintf(stderr,"line too long\n");
	    goto reset;
	}
	*pos++ = ch;
    }
}


static jmp_buf env;


static void handle_ctrl_c(int sig)
{
    (void) write(1,"\n",1);
    longjmp(env,1);
}


static void handle_commands(void)
{
    struct sigaction act;
    sigset_t set;

    if (setjmp(env)) unwind_stack();
    if (shmod_catch_signals) {
	if (sigemptyset(&set) < 0) {
	    perror("sigemptyset");
	    exit(1);
	}
	act.sa_handler = &handle_ctrl_c;
	act.sa_mask = set;
	act.sa_flags = 0;
	if (sigaction(SIGINT,&act,NULL) < 0) {
	    perror("sigaction");
	    exit(1);
	}
	if (sigaddset(&set,SIGINT) < 0) {
	    perror("sigaddset");
	    exit(1);
	}
	if (sigprocmask(SIG_UNBLOCK,&set,NULL) < 0) {
	    perror("sigprocmask");
	    exit(1);
	}
    }
    while (1) do_handle_command();
}


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s [ -c command ]\n",name);
    exit(1);
}


#ifdef STANDALONE
#define sh_mod main
#endif


int sh_mod(int argc,char **argv)
{
    char *command = NULL;

    if (argc != 1) {
	if (argc != 3 || strcmp(argv[1],"-c")) usage(argv[0]);
	command = argv[2];
    }
    push_input(&in_eof,NULL);
    if (command) push_input(&in_string,command);
    else {
	FILE *file;

	push_input(&in_file,stdin);
	file = fopen(RC_FILE,"r");
	if (file) push_input(&in_file,file);
    }
    handle_commands();
    return 0; /* just in case */
}
