/*
 * tcpcp.c - TCP connection passing high-level API
 *
 * Written 2002 by Werner Almesberger
 * Distributed under the LGPL.
 */


#include <stddef.h>
#include <inttypes.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "linux/tcpcp.h"

#include "tcpcp.h"


/* ----- Interface to low-level API ---------------------------------------- */


static int tcp_max_ici_size(int s,int *size)
{
    int size_size = sizeof(*size);

    return getsockopt(s,SOL_TCP,TCP_MAXICISIZE,size,&size_size);
}


static int tcp_get_ici(int s,void *ici,int size)
{
    return getsockopt(s,SOL_TCP,TCP_ICI,ici,&size);
}


static int tcp_set_ici(int s,const void *ici,int size)
{
    return setsockopt(s,SOL_TCP,TCP_ICI,ici,size);
}


/* ----- Public functions -------------------------------------------------- */


void *tcpcp_get(int s)
{
    int size,saved_errno;
    void *ici;

    if (tcp_max_ici_size(s,&size) < 0) return NULL;
    ici = malloc(size);
    if (!ici) return NULL;
    if (!tcp_get_ici(s,ici,size)) return ici;
    saved_errno = errno;
    free(ici);
    errno = saved_errno;
    return NULL;
}


int tcpcp_size(const void *ici)
{
    const struct tcpcp_ici *_ici = ici;

    return ntohl(_ici->ici_length);
}


int tcpcp_create(const void *ici)
{
    int s,saved_errno;

    s = socket(PF_INET,SOCK_STREAM,0);
    if (s < 0) return s;
    if (!tcp_set_ici(s,ici,tcpcp_size(ici))) return s;
    saved_errno = errno;
    (void) close(s);
    errno = saved_errno;
    return -1;
}


int tcpcp_set_cong(void *ici,int cong_mode)
{
    struct tcpcp_ici *_ici = ici;

    switch (cong_mode) {
	case TCPCP_CONG_DEFAULT:
	case TCPCP_CONG_NEW:
	    _ici->v.flags &= ~htons(TCPCP_ICIF_USEPERF);
	    break;
	case TCPCP_CONG_KEEP:
	    _ici->v.flags |= htons(TCPCP_ICIF_USEPERF);
	    break;
	default:
	    return -1;
    }
    return 0;
}
